/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * \package    RobotAPI::gui-plugins::SkillManagerMonitorWidgetController
 * \author     Raphael Grimm ( raphael dot grimm at kit dot edu )
 * \date       2020
 * \copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "SkillManagerMonitorWidgetController.h"

#include <string>

#include <QClipboard>
#include <QDoubleSpinBox>
#include <QGridLayout>
#include <QTextBrowser>

#include <RobotAPI/libraries/aron/converter/json/NLohmannJSONConverter.h>
#include <RobotAPI/libraries/skills/core/Skill.h>
#include <RobotAPI/libraries/skills/core/SkillExecutionRequest.h>

#include "aronTreeWidget/Data.h"
#include "aronTreeWidget/modal/text/AronTreeWidgetTextInputModalController.h"
#include "aronTreeWidget/visitors/AronTreeWidgetConverter.h"
#include "aronTreeWidget/visitors/AronTreeWidgetCreator.h"
#include "aronTreeWidget/visitors/AronTreeWidgetModalCreator.h"
#include "aronTreeWidget/widgets/SkillDescriptionWidget.h"

namespace armarx
{
    QPointer<QDialog>
    SkillManagerMonitorWidgetController::getConfigDialog(QWidget* parent)
    {
        if (!dialog)
        {
            dialog = new SimpleConfigDialog(parent);
            dialog->addProxyFinder<skills::dti::SkillMemoryInterfacePrx>(
                "SkillMemory", "", "SkillMem*");
        }
        return qobject_cast<SimpleConfigDialog*>(dialog);
    }

    void
    SkillManagerMonitorWidgetController::configured()
    {
        observerName = dialog->getProxyName("SkillMemory");
    }

    void
    SkillManagerMonitorWidgetController::loadSettings(QSettings* settings)
    {
        observerName = settings->value("SkillMemory", "SkillMemory").toString().toStdString();
    }

    void
    SkillManagerMonitorWidgetController::saveSettings(QSettings* settings)
    {
        settings->setValue("SkillMemory", QString::fromStdString(observerName));
    }
} // namespace armarx

// Others
namespace armarx
{
    SkillExecutionInfoTreeWidgetItem*
    SkillExecutionInfoTreeWidgetItem::SearchRecursiveForMatch(
        SkillExecutionInfoTreeWidgetItem* haystack,
        const skills::SkillExecutionID& needle)
    {
        if (!haystack)
        {
            return nullptr;
        }

        if (needle == haystack->executionId)
        {
            return haystack;
        }
        for (int i = 0; i < haystack->childCount(); ++i)
        {
            auto el = static_cast<SkillExecutionInfoTreeWidgetItem*>(haystack->child(i));
            return SkillExecutionInfoTreeWidgetItem::SearchRecursiveForMatch(el, needle);
        }
        return nullptr;
    }

    SkillManagerMonitorWidgetController::SkillManagerMonitorWidgetController()
    {
        widget.setupUi(getWidget());

        double default_hz = 2;
        widget.doubleSpinBoxUpdateFreq->setValue(default_hz);
        widget.doubleSpinBoxUpdateFreq->setMinimum(0.5);
        widget.doubleSpinBoxUpdateFreq->setMaximum(20);
        widget.doubleSpinBoxUpdateFreq->setSingleStep(0.5);
        widget.doubleSpinBoxUpdateFreq->setSuffix(" Hz");

        skillDescriptionWidget = new SkillDescriptionWidget();
        widget.skillDescription->parentWidget()->layout()->replaceWidget(widget.skillDescription,
                                                                         skillDescriptionWidget);
        widget.skillDescription = skillDescriptionWidget;


        refreshSkillsResultTimer = new QTimer(this);
        updateTimerFrequency();
        refreshSkillsResultTimer->start();

        connect(widget.doubleSpinBoxUpdateFreq,
                &QDoubleSpinBox::editingFinished,
                this,
                &SkillManagerMonitorWidgetController::updateTimerFrequency);
        connect(refreshSkillsResultTimer,
                &QTimer::timeout,
                this,
                &SkillManagerMonitorWidgetController::refreshSkillsAndExecutions);

        connect(widget.pushButtonCopy,
                &QPushButton::clicked,
                this,
                &SkillManagerMonitorWidgetController::copyCurrentConfig);
        connect(widget.pushButtonPaste,
                &QPushButton::clicked,
                this,
                &SkillManagerMonitorWidgetController::pasteCurrentConfig);

        connect(widget.pushButtonExecuteSkill,
                &QPushButton::clicked,
                this,
                &SkillManagerMonitorWidgetController::executeSkill);
        //        connect(widget.pushButtonStopSkill,
        //                &QPushButton::clicked,
        //                this,
        //                &SkillManagerMonitorWidgetController::stopSkill);

        connect(widget.treeWidgetSkills,
                &QTreeWidget::currentItemChanged,
                this,
                &SkillManagerMonitorWidgetController::skillSelectionChanged);

        connect(widget.pushButtonRefreshNow,
                &QPushButton::clicked,
                this,
                &SkillManagerMonitorWidgetController::refreshSkills);
    }

    SkillManagerMonitorWidgetController::~SkillManagerMonitorWidgetController()
    {
    }

    void
    SkillManagerMonitorWidgetController::onInitComponent()
    {
        usingProxy(observerName);
    }

    void
    SkillManagerMonitorWidgetController::onConnectComponent()
    {
        widget.groupBoxSkills->setTitle(QString::fromStdString(observerName));
        widget.treeWidgetSkillDetails->setEditTriggers(
            QAbstractItemView::EditTrigger::NoEditTriggers);
        widget.treeWidgetSkillDetails->setColumnHidden(3, true);

        getProxy(memory, observerName, 1000);
        connected = true;
    }

    void
    SkillManagerMonitorWidgetController::onDisconnectComponent()
    {
        connected = false;
        memory = nullptr;

        // reset all
        skills.clear();
        widget.treeWidgetSkills->clear();
        widget.treeWidgetSkillDetails->clear();
        skillsArgumentsTreeWidgetItem = nullptr;
        selectedSkill.skillId.providerId->providerName = "";
        selectedSkill.skillId.skillName = "";
    }

    void
    SkillManagerMonitorWidgetController::updateTimerFrequency()
    {
        int f = static_cast<int>(std::round(1000 / widget.doubleSpinBoxUpdateFreq->value()));
        refreshSkillsResultTimer->setInterval(f);
    }

    void
    SkillManagerMonitorWidgetController::refreshSkillsAndExecutions()
    {
        if (widget.checkBoxAutoUpdate->isChecked())
        {
            refreshSkills();
            refreshExecutions();
        }
    }

    void
    SkillManagerMonitorWidgetController::refreshSkills()
    {
        if (!memory)
        {
            // check if null
            return;
        }

        /* CHECK OWN SKILLS LIST */
        // remove non-existing ones
        try
        {
            std::scoped_lock l(updateMutex);

            auto managerSkills = memory->getSkillDescriptions();

            // completely recreate internal skills map
            skills.clear();
            for (const auto& [sid, desc] : managerSkills)
            {
                auto description = skills::SkillDescription::FromIce(desc);
                auto skillId = skills::SkillID::FromIce(sid);
                auto providerId = skillId.providerId.value_or(
                    skills::ProviderID{.providerName = "UNKNOWN PROVIDER NAME"});

                ARMARX_CHECK(skillId.isFullySpecified());

                auto& providedSkillsMap = skills[providerId]; // create new if not existent
                providedSkillsMap.insert({skillId, description});
            }

            // update tree view. Remove non-existing elements
            int i = 0;
            while (i < widget.treeWidgetSkills->topLevelItemCount())
            {
                auto* providerItem = widget.treeWidgetSkills->topLevelItem(i);
                auto providerName = providerItem->text(0).toStdString();
                skills::ProviderID providerId{.providerName = providerName};

                if (skills.find(providerId) == skills.end())
                {
                    providerItem = nullptr; // reset
                    auto remove = widget.treeWidgetSkills->takeTopLevelItem(i);
                    delete remove;
                    continue;
                }

                ++i;
                auto& providedSkills = skills.at(providerId);

                int j = 0;
                while (j < providerItem->childCount())
                {
                    auto* skillItem = providerItem->child(j);
                    auto skillName = skillItem->text(0).toStdString();

                    skills::SkillID skillId{.providerId =
                                                skills::ProviderID{.providerName = providerName},
                                            .skillName = skillName};

                    if (providedSkills.find(skillId) == providedSkills.end())
                    {
                        skillItem = nullptr;
                        auto remove = providerItem->takeChild(j);
                        delete remove;
                        continue;
                    }

                    ++j;
                }
            }

            // update tree view. Add new elements
            for (const auto& [providerId, providedSkills] : skills)
            {
                QTreeWidgetItem* providerItem = nullptr;
                for (int i = 0; i < widget.treeWidgetSkills->topLevelItemCount(); ++i)
                {
                    auto el = widget.treeWidgetSkills->topLevelItem(i);
                    auto providerName = el->text(0).toStdString();
                    skills::ProviderID elProviderId{.providerName = providerName};

                    if (providerId == elProviderId)
                    {
                        providerItem = el;
                        break;
                    }
                }

                if (!providerItem)
                {
                    providerItem = new QTreeWidgetItem(widget.treeWidgetSkills);
                    providerItem->setText(0, QString::fromStdString(providerId.providerName));
                }

                for (const auto& [skillId, skill] : providedSkills)
                {
                    QTreeWidgetItem* skillItem = nullptr;
                    for (int i = 0; i < providerItem->childCount(); ++i)
                    {
                        auto el = providerItem->child(i);
                        auto skillName = el->text(0).toStdString();
                        skills::SkillID elSkillId{providerId, skillName};

                        if (skillId == elSkillId)
                        {
                            skillItem = el;
                            break;
                        }
                    }

                    if (!skillItem)
                    {
                        skillItem = new SkillInfoTreeWidgetItem(skill, providerItem);
                        skillItem->setText(0, QString::fromStdString(skillId.skillName));
                    }
                }
            }
        }
        catch (...)
        {
            // perhaps the manager died during the method?
        }
    }

    void
    SkillManagerMonitorWidgetController::refreshExecutions()
    {
        static std::map<skills::SkillStatus, std::string> ExecutionStatus2String = {
            // Terminating
            {skills::SkillStatus::Aborted, "Aborted"},
            {skills::SkillStatus::Failed, "Failed"},
            {skills::SkillStatus::Succeeded, "Succeeded"},

            // Others
            {skills::SkillStatus::Constructing, "Under construction"},
            {skills::SkillStatus::Running, "Running"},
            {skills::SkillStatus::Initializing, "Initializing"},
            {skills::SkillStatus::Preparing, "Preparing"}};

        if (!memory)
        {
            // check if null
            return;
        }

        try
        {
            std::scoped_lock l(updateMutex);

            auto currentManagerStatuses =
                memory
                    ->getSkillExecutionStatuses(); // we assume that there are no more than 100 new skills..

            for (const auto& [k, v] : currentManagerStatuses)
            {
                auto executionId = skills::SkillExecutionID::FromIce(k);
                auto statusUpdate = skills::SkillStatusUpdate::FromIce(v);

                SkillExecutionInfoTreeWidgetItem* found = nullptr;
                for (int i = 0; i < widget.treeWidgetSkillExecutions->topLevelItemCount(); ++i)
                {
                    auto c = static_cast<SkillExecutionInfoTreeWidgetItem*>(
                        widget.treeWidgetSkillExecutions->topLevelItem(i));

                    found =
                        SkillExecutionInfoTreeWidgetItem::SearchRecursiveForMatch(c, executionId);

                    if (found)
                    {
                        // update values
                        found->setText(3,
                                       QString::fromStdString(
                                           statusUpdate.hasBeenConstructed() ? " yes " : " no "));
                        found->setText(4,
                                       QString::fromStdString(
                                           statusUpdate.hasBeenInitialized() ? " yes " : " no "));
                        found->setText(5,
                                       QString::fromStdString(
                                           statusUpdate.hasBeenPrepared() ? " yes " : " no "));
                        found->setText(6,
                                       QString::fromStdString(
                                           statusUpdate.hasBeenRunning() ? " yes " : " no "));
                        found->setText(7,
                                       QString::fromStdString(
                                           statusUpdate.hasBeenTerminated() ? " yes " : " no "));
                        break;
                    }
                }

                if (!found)
                {
                    // TODO: Sort to executor!
                    auto item = new SkillExecutionInfoTreeWidgetItem(
                        executionId, widget.treeWidgetSkillExecutions);

                    item->setText(0,
                                  QString::fromStdString(
                                      executionId.executionStartedTime.toDateTimeString()));
                    item->setText(1, QString::fromStdString(executionId.executorName));
                    item->setText(2, QString::fromStdString(executionId.skillId.toString()));
                    item->setText(3,
                                  QString::fromStdString(
                                      statusUpdate.hasBeenConstructed() ? " yes " : " no "));
                    item->setText(4,
                                  QString::fromStdString(
                                      statusUpdate.hasBeenInitialized() ? " yes " : " no "));
                    item->setText(
                        5,
                        QString::fromStdString(statusUpdate.hasBeenPrepared() ? " yes " : " no "));
                    item->setText(
                        6,
                        QString::fromStdString(statusUpdate.hasBeenRunning() ? " yes " : " no "));
                    item->setText(7,
                                  QString::fromStdString(
                                      statusUpdate.hasBeenTerminated() ? " yes " : " no "));
                }
            }
        }
        catch (...)
        {
            // perhaps the manager died during the method?
        }
    }

    void
    SkillManagerMonitorWidgetController::executeSkill()
    {
        if (not selectedSkill.skillId.isFullySpecified())
        {
            return;
        }

        std::scoped_lock l(updateMutex);

        auto providerId = *selectedSkill.skillId.providerId;
        const auto& skillDescriptions = skills.at(providerId);
        if (skillDescriptions.find(selectedSkill.skillId) == skillDescriptions.end())
        {
            return;
        }

        auto data = getConfigAsAron();

        char hostname[HOST_NAME_MAX];
        gethostname(hostname, HOST_NAME_MAX);

        skills::SkillExecutionRequest req{selectedSkill.skillId,
                                          "Skills.Manager GUI (hostname: " + std::string(hostname) +
                                              ")",
                                          data};

        ARMARX_CHECK(selectedSkill.skillId.isFullySpecified()); // sanity check
        ARMARX_IMPORTANT << "Executing skill from GUI: " << selectedSkill.skillId << ".";
        // Note that we execute the skill in a seperate thread so that the GUI thread does not freeze.
        memory->begin_executeSkill(req.toManagerIce());
    }

    void
    SkillManagerMonitorWidgetController::stopSkill()
    {
        //        std::scoped_lock l(updateMutex);
        //        if (selectedSkill.skillId.isFullySpecified())
        //        {
        //            return;
        //        }

        //        const auto& skillDescriptions = skills.at(*selectedSkill.skillId.providerId);
        //        if (!skillDescriptions.count(selectedSkill.skillId.skillName))
        //        {
        //            return;
        //        }

        //        ARMARX_INFO << "Stopping skill from GUI: " << selectedSkill.skillId;

        //        manager->abortSkill(selectedSkill.skillId);
    }

    void
    SkillManagerMonitorWidgetController::skillSelectionChanged(QTreeWidgetItem* current,
                                                               QTreeWidgetItem*)
    {
        std::scoped_lock l(updateMutex);
        widget.groupBoxSkillDetails->setEnabled(false);

        if (!current)
        {
            // gui has died?
            return;
        }

        if (!current->parent())
        {
            // no parent available. Perhaps provider clicked? Reset selected skill.
            selectedSkill.skillId.providerId->providerName = "";
            selectedSkill.skillId.skillName = "";
            return;
        }

        auto c = static_cast<SkillInfoTreeWidgetItem*>(current);
        auto skillDescription = c->skillDescription;

        if (selectedSkill.skillId == skillDescription.skillId)
        {
            // no change
            return;
        }

        selectedSkill.skillId = skillDescription.skillId;

        // setup groupBox
        widget.groupBoxSkillDetails->setTitle(
            QString::fromStdString(selectedSkill.skillId.toString()));
        widget.groupBoxSkillDetails->setEnabled(true);

        // setup table view
        widget.treeWidgetSkillDetails->clear();
        aronTreeWidgetController = nullptr;
        skillsArgumentsTreeWidgetItem = nullptr;

        // We assert that the skill exists
        ARMARX_CHECK(skills.count(*selectedSkill.skillId.providerId) > 0);
        ARMARX_CHECK(skills.at(*selectedSkill.skillId.providerId).count(selectedSkill.skillId) > 0);
        auto skillDesc = skills.at(*selectedSkill.skillId.providerId).at(selectedSkill.skillId);

        skillDescriptionWidget->setSkillDescription(skillDesc);

        // select root profile
        widget.comboBoxProfiles->setCurrentIndex(0);

        // remove any profile
        while (widget.comboBoxProfiles->count() > 1)
        {
            widget.comboBoxProfiles->removeItem(1);
        }

        // add new profiles for this skill
        // TODO: Where stored?

        skillsArgumentsTreeWidgetItem = new QTreeWidgetItem(widget.treeWidgetSkillDetails,
                                                            {QString::fromStdString("Arguments")});
        auto aron_args = skillDesc.parametersType;
        auto default_args_of_profile = skillDesc.rootProfileDefaults;

        aronTreeWidgetController =
            std::make_shared<AronTreeWidgetController>(widget.treeWidgetSkillDetails,
                                                       skillsArgumentsTreeWidgetItem,
                                                       aron_args,
                                                       default_args_of_profile);

        // automatically expand args
        skillsArgumentsTreeWidgetItem->setExpanded(true);
    }

    aron::data::DictPtr
    SkillManagerMonitorWidgetController::getConfigAsAron() const
    {
        // create argument aron (if there is an accepted type set)
        if (aronTreeWidgetController)
        {
            return aronTreeWidgetController->convertToAron();
        }
        return nullptr;
    }

    void
    SkillManagerMonitorWidgetController::copyCurrentConfig()
    {
        auto data = getConfigAsAron();
        if (!data)
        {
            return;
        }

        auto json = aron::data::converter::AronNlohmannJSONConverter::ConvertToNlohmannJSON(data);
        QClipboard* clipboard = QApplication::clipboard();
        clipboard->setText(QString::fromStdString(json.dump(2)));
    }

    void
    SkillManagerMonitorWidgetController::pasteCurrentConfig()
    {
        QClipboard* clipboard = QApplication::clipboard();
        std::string s = clipboard->text().toStdString();
        nlohmann::json json = nlohmann::json::parse(s);
        auto data =
            aron::data::converter::AronNlohmannJSONConverter::ConvertFromNlohmannJSONObject(json);

        if (!aronTreeWidgetController)
        {
            return;
        }

        aronTreeWidgetController->setFromAron(data);
    }

    void
    SkillManagerMonitorWidgetController::resetCurrentConfig()
    {
        // TODO
    }

} // namespace armarx
