// Header
#include "DiskStorageMixin.h"

// STD/STL
#include <algorithm>
#include <fstream>
#include <iostream>

// Simox
#include <SimoxUtility/algorithm/string.h>

// ArmarX
#include <ArmarXCore/core/exceptions/LocalException.h>
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/time/TimeUtil.h>

namespace armarx::armem::server::ltm::detail::mixin
{
    DiskMemoryItemMixin::DiskMemoryItemMixin(const Path& memoryParentPath,
                                             const std::string& exportName,
                                             const armem::MemoryID& id) :
        memoryBasePath(memoryParentPath), exportName(exportName), _id(id)
    {
    }

    void
    DiskMemoryItemMixin::setMemoryBasePath(const Path& n)
    {
        memoryBasePath = n;
    }

    void
    DiskMemoryItemMixin::setMixinExportName(const std::string& n)
    {
        exportName = n;
    }

    void
    DiskMemoryItemMixin::setMixinMemoryID(const MemoryID& n)
    {
        ARMARX_CHECK_NOT_EMPTY(_id.memoryName) << " The full id was: " << _id.str();

        _id = n;
    }

    void
    DiskMemoryItemMixin::configureMixin(const nlohmann::json& json)
    {
        if (json.find("DiskMemory.memoryParentPath") != json.end())
        {
            memoryBasePath = Path(json.at("DiskMemory.memoryParentPath"));
        }
    }

    Path
    DiskMemoryItemMixin::getMemoryBasePath() const
    {
        return memoryBasePath;
    }

    Path
    DiskMemoryItemMixin::getFullPath() const
    {
        auto p = getMemoryBasePath() / exportName;
        return util::fs::toPath(p, _id);
    }

    bool
    DiskMemoryItemMixin::memoryBasePathExists() const
    {
        return util::fs::directoryExists(getMemoryBasePath());
    }

    bool
    DiskMemoryItemMixin::fullPathExists() const
    {
        auto p = getFullPath();
        return util::fs::directoryExists(p);
    }

    bool
    DiskMemoryItemMixin::fileExists(const std::string& filename) const
    {
        auto p = getFullPath() / filename;
        return util::fs::fileExists(p);
    }

    void
    DiskMemoryItemMixin::ensureMemoryBasePathExists(bool createIfNotExistent) const
    {
        util::fs::ensureDirectoryExists(getMemoryBasePath(), createIfNotExistent);
    }

    void
    DiskMemoryItemMixin::ensureFullPathExists(bool createIfNotExistent) const
    {
        auto p = getFullPath();
        util::fs::ensureDirectoryExists(p, createIfNotExistent);
    }

    void
    DiskMemoryItemMixin::ensureFileExists(const std::string& filename,
                                          bool createIfNotExistent) const
    {
        auto p = getFullPath() / filename;
        util::fs::ensureFileExists(p, createIfNotExistent);
    }

    void
    DiskMemoryItemMixin::writeDataToFile(const std::string& filename,
                                         const std::vector<unsigned char>& data) const
    {
        auto p = getFullPath() / filename;
        util::fs::writeDataToFile(p, data);
    }

    std::vector<unsigned char>
    DiskMemoryItemMixin::readDataFromFile(const std::string& filename) const
    {
        auto p = getFullPath() / filename;
        return util::fs::readDataFromFile(p);
    }

    std::vector<Path>
    DiskMemoryItemMixin::getAllDirectories() const
    {
        auto p = getFullPath();
        return util::fs::getAllDirectories(p);
    }

    std::vector<Path>
    DiskMemoryItemMixin::getAllFiles() const
    {
        auto p = getFullPath();
        return util::fs::getAllFiles(p);
    }
} // namespace armarx::armem::server::ltm::detail::mixin
